// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

#ifndef _checker_H
#define _checker_H   1

#include "grid.h"
#include <time.h>
#include <vector>
#include "int64.h"
#include "stopwatch.h"

using namespace std;

// Only 64 * NUM_PACKS additional unavoidable sets are allowed
// ('additional' refers to those unavodables that are not members
// of the maximal collection of pairwise disjoint unavoidable sets)
// 1 = 64 add'l unav'bles, 2 = 128, 3 = 192, 4 = 256, 5 = 320
#ifndef NUM_PACKS
   #define NUM_PACKS 3 // use up to 192 unavoidables by default
#endif

struct BitMask64
   {
   int64 aBits[ NUM_PACKS ];
   };
typedef BitMask64 bitmask_t;

#if NUM_PACKS == 1
   #define PROG_EXECUTABLE_NAME "checker64"
   #define XOR_BITMASKS( index )    \
         m_SetMask.aBits[ 0 ] |= m_aHittingMasks[ (index) ].aBits[ 0 ];
#elif NUM_PACKS == 2
   #define PROG_EXECUTABLE_NAME "checker128"
   #define XOR_BITMASKS( index )    \
         m_SetMask.aBits[ 0 ] |= m_aHittingMasks[ (index) ].aBits[ 0 ];    \
         m_SetMask.aBits[ 1 ] |= m_aHittingMasks[ (index) ].aBits[ 1 ];
#elif NUM_PACKS == 3
   #define PROG_EXECUTABLE_NAME "checker192"
   #define XOR_BITMASKS( index )    \
         m_SetMask.aBits[ 0 ] |= m_aHittingMasks[ (index) ].aBits[ 0 ];    \
         m_SetMask.aBits[ 1 ] |= m_aHittingMasks[ (index) ].aBits[ 1 ];    \
         m_SetMask.aBits[ 2 ] |= m_aHittingMasks[ (index) ].aBits[ 2 ];
#elif NUM_PACKS == 4
   #define PROG_EXECUTABLE_NAME "checker256"
   #define XOR_BITMASKS( index )    \
         m_SetMask.aBits[ 0 ] |= m_aHittingMasks[ (index) ].aBits[ 0 ];    \
         m_SetMask.aBits[ 1 ] |= m_aHittingMasks[ (index) ].aBits[ 1 ];    \
         m_SetMask.aBits[ 2 ] |= m_aHittingMasks[ (index) ].aBits[ 2 ];    \
         m_SetMask.aBits[ 3 ] |= m_aHittingMasks[ (index) ].aBits[ 3 ];
#elif NUM_PACKS == 5
   #define PROG_EXECUTABLE_NAME "checker320"
   #define XOR_BITMASKS( index )    \
         m_SetMask.aBits[ 0 ] |= m_aHittingMasks[ (index) ].aBits[ 0 ];    \
         m_SetMask.aBits[ 1 ] |= m_aHittingMasks[ (index) ].aBits[ 1 ];    \
         m_SetMask.aBits[ 2 ] |= m_aHittingMasks[ (index) ].aBits[ 2 ];    \
         m_SetMask.aBits[ 3 ] |= m_aHittingMasks[ (index) ].aBits[ 3 ];    \
         m_SetMask.aBits[ 4 ] |= m_aHittingMasks[ (index) ].aBits[ 4 ];
#else
   #error NUM_PACKS must be in the range 1-5.
#endif


struct Cell
   {
   int m_iSet;   // index of the unavoidable set
   int m_iCell;  // index of the cell within the unavoidable set
   int m_Coords; // coordinates of this cell within the grid, e.g. 24 = row 2, column 4
   };


// stores an n-puzzle (or a k-pseudo-n-puzzle, where k is the # of completions)
class Puzzle
   {
   public:
      Puzzle( int cClues, Cell* const achCells, int cCompletions ) : m_cCompletions( cCompletions ),
         m_cClues( cClues ), m_achClues( new char[ cClues ] )
         {
         for ( int i = 0;  i < m_cClues;  i++ )
            m_achClues[ i ] = static_cast< char >( achCells[ i ].m_Coords );
         qsort( m_achClues, m_cClues, sizeof( char ), Puzzle::CharComp );
         }

      Puzzle( char szPuzzle[ 81 ], int cCompletions ) : m_cCompletions( cCompletions )
         {
         // first count # of clues
         m_cClues = 0;
         for ( int i = 0;  i < 81;  i++ )
            if ( szPuzzle[ i ] != '0' )
               m_cClues++;
         // now record the coordiates of the clues
         m_achClues = new char[ m_cClues ];
         int count = 0;
         for ( int i = 0;  i < 81;  i++ )
            if ( szPuzzle[ i ] != '0' )
               m_achClues[ count++ ] = static_cast< char >( (i/9 + 1) * 10 + (i%9 + 1) );
         }

      Puzzle( Puzzle const& rhs )
         {
         m_achClues = NULL;
         *this = rhs;
         }

      static int CharComp( const void* pch1, const void* pch2 )
         {
         return ( *((char*)pch1) - *((char*)pch2) );
         }

      // to allow sorting -- sort first by the number of completions and then
      // by the coordinates of the clues (that's why we need the clues to be sorted)
      bool operator<( Puzzle const& rhs ) const
         {
         if ( m_cCompletions < rhs.m_cCompletions  ||
              m_cCompletions == rhs.m_cCompletions  &&
              memcmp( m_achClues, rhs.m_achClues, m_cClues * sizeof(char) ) < 0 )
            return true;
         else
            return false;
         }

      bool operator==( Puzzle const& rhs ) const
         {
         return ( m_cCompletions == rhs.m_cCompletions  &&
                  0 == memcmp( m_achClues, rhs.m_achClues, m_cClues * sizeof( char )) );
         }

      Puzzle const& operator=( Puzzle const& rhs )
         {
         delete[] m_achClues;
         m_cCompletions = rhs.m_cCompletions;
         m_cClues = rhs.m_cClues;
         m_achClues = new char[ m_cClues ];
         memcpy( m_achClues, rhs.m_achClues, m_cClues * sizeof( char ) );
         return *this;
         }

      void PrintLong( ostream& out, Grid const& grid ) const
         {
         out << m_cCompletions << " completions: ";
         Print( out, grid );
         }

      void Print( ostream& out, Grid const& grid ) const
         {
         char sz[ 82 ];
         memset( sz, '0', 81 );
         sz[ 81 ] = '\0';
         for ( int i = 0;  i < m_cClues;  i++ )
            {
            int ich = Coords2Idx( m_achClues[ i ] );
            sz[ ich ] = grid.GetGrid()[ ich ];
            }
         out << sz << endl;
         }

      int GetCompletions() const { return m_cCompletions; }

      ~Puzzle() { delete[] m_achClues; }

   private:
      int   m_cCompletions;
      int   m_cClues;
      char* m_achClues;
   };

ostream& operator<<( ostream&, Puzzle const& );

class ClueChecker
   {
   public:
      ClueChecker( Grid const& grid, char const* pszPuzzFile, int cClues, int nProbability );
      ~ClueChecker();
      void DoChecking( int cFirstClues = 0 );
      void DoRandomChecking();
      // bogus assignment operator to avoid compiler warning (default assignment
      // operator cannot be generated because some member variables are constant)
      void operator=( ClueChecker const& ) { printf( "Not allowed." ); }
   private:
      void GenerateTuples();
      void ProcessTuple();
      void SavePuzzles();
      inline void AddClue( int iSet, int iCell );
      void SavePseudoPuzzles( char const* pszFilename );
      void ComputeHittingMasks();
      bool HasRunBefore();
      void PrintTime( int seconds ) const
         {
         int days    = seconds / 86400;
         seconds %= 86400;
         int hours   = seconds / 3600;
         seconds %= 3600;
         int minutes = seconds / 60;
         printf( "%dd %02dh %02dm", days, hours, minutes );
         }

      inline void GenerateAndCheckPuzzle();
      inline bool AreAllUnavoidableSetsHit( int& iNotHit );
      inline void DrawRemainingClues( int iStart = 0 );

   private:
      Grid const& m_grid;
      char const* m_pszPuzzFile;
      ofstream*   m_posPuzzFile;
      ofstream*   m_posPseuPuzzFile;
      int const   m_cClues;
      int const   m_nProbability;
      Cell* const m_pTuple;
      int         m_iIndex;
      int         m_cProgressRounds;
      int         m_cTotalRounds;
      int         m_cRounds;
      int64       m_cPuzzles;  // total # of puzzles generated during enumeration
      int64       m_cPuzzlesSolved; // # of puzzles that were actually run through the solver
      int         m_cUniques; // # of puzzles found having a unique completion
      int         m_cPseudos; // # of puzzles found having (exactly) 2 completions
      char* const m_szPuzzle;
      int*        m_piCandidateClues;
      int         m_cCandidateClues;
      time_t      m_tmStart;
      bitmask_t   m_SetMask;
      bitmask_t   m_aHittingMasks[ 100 ]; //11-99
      // for measuring time spent in solver
      Stopwatch   m_StopwatchSolver; 
      Stopwatch   m_Stopwatch;
      bool        m_bRun;      
      int         m_aiDeadClues[ 3 ]; // clues that do not have to be considered when generating puzzles
      int*        m_aiLowestBit; // array: {0,1,...,65535} -> lowest zero bit in binary expansion


      // arrays so we can keep track of n-puzzles and pseudo-n-puzzles found
      // An n-puzzle is a puzzle of n clues that has a unique solution,
      // a pseudo-n-puzzle is a puzzle of n clues that has exactly two solutions.
      vector< Puzzle > m_vcUniquePuzzles;
      vector< Puzzle > m_vcPseudoPuzzles;
   };

#endif
